
module TypeChecker.Values where

import Syntax.Internal
import Syntax.Pretty
import Syntax.Position

import Utils.Map
import Utils.Pairs
import Utils.Debug

import {-# SOURCE #-} TypeChecker.UnfoldConstants -- REMOVE_IF_HADDOCK

{- REMOVE_IF_HADDOCK
-- | Cyclic module dependency with "TypeChecker.UnfoldConstants". The real function
--   is 'TypeChecker.UnfoldConstants.unfoldConstants'.
unfoldConstants :: Value -> Value
REMOVE_IF_HADDOCK -}

-- | Compute the type of a value.
valueType :: Value -> Type
valueType v =
    case v of
	VId x t			-> t
	VDef x (Definition v)	-> valueType v
	VDef x (Axiom t)	-> t
	VPi v cl t		-> t
	VLam cl t		-> t
	VSort (VSUniv u)	-> VSort $ VSUniv $ u + 1
	VSort VSProp		-> VSort $ VSUniv 1
	VTop t			-> t
	VApp v w		-> valueType v `piApp` w

{- | Give an argument to a function type.

>  piApp ((x :: A) -> B) w = B[w/x]
 
Raises an internal error if the first argument does not have a 
function type. -}
piApp :: Type -> Value -> Type
piApp v w =
    case unfoldConstants v of
	VPi _ f _   -> clApp f w
	t	    -> error $ "Internal error in 'piApp': "
				++ show v
				++ " should be a function but is "
				++ show t
				++ " in the application to "
				++ show w

vApps :: Value -> [Value] -> Value
vApps = foldl vApp

vApp :: Value -> Value -> Value
vApp (VLam f _) w   = clApp f w
vApp (VTop t) w	    = VTop (t `piApp` w)
vApp v w	    = VApp v w

-- | Performs applications and projections that might have been delayed
--   by tagging.
class Normalize a where
    normalize :: a -> a

instance Normalize a => Normalize [a] where
    normalize = map normalize

instance (Normalize a, Normalize b) => Normalize (a,b) where
    normalize = normalize -*- normalize

instance Normalize Value where
    normalize v =
	case v of
	    VId x t		-> VId x (normalize t)
	    VDef x d		-> VDef x d
	    VApp v w		-> vApp (normalize v) (normalize w)
	    VPi v f t		-> VPi (normalize v) (normalize f) (normalize t)
	    VLam f t		-> VLam (normalize f) (normalize t)
	    VSort s		-> VSort s
	    VTop t		-> VTop (normalize t)
	    VTag v		-> normalize v	-- TODO: this can't be right!

instance Normalize Closure where
    normalize (Cl x f) = Cl x $ \v -> normalize (f v)

