{-# OPTIONS -fglasgow-exts -fallow-overlapping-instances #-}

module TypeChecker.Values.Refresh where

import Control.Monad.Trans
import Data.IORef

import Syntax.Internal
import Syntax.Values
import TypeChecker.Context
import TypeChecker.Monad
import {-# SOURCE #-} TypeChecker.Values
import Utils.Monads

-- | Only refresh top-level metas
refreshTop :: Value -> TC Value
refreshTop v =
    case vAppView v of
	VAVMeta m vs	->
	    do  mv <- liftIO $ readIORef m
		case mv of
		    Open	    -> return v
		    Instantiated v  -> refreshTop (v `vApps` vs)
	_ -> return v

refreshSpine :: Value -> TC Value
refreshSpine v =
    do	v' <- refreshTop v
	case appView v' of
	    v:vs    -> vApps v <$> mapM refreshTop vs
	    _	    -> error "impossible: TypeChecker.Values.Refresh.refreshSpine"
    where
	appView v =
	    case vView v of
		VApp v w    -> appView v ++ [w]
		_	    -> [v]

class Refresh a where
    refresh :: a -> TC a

instance Refresh Value where
    refresh v =
	case vAppView v of
	    VAVId x vs	    -> foldl vApp (vId x) <$> refresh vs
	    VAVBId n vs	    -> foldl vApp (vBId n) <$> refresh vs
	    VAVMeta m vs    ->
		do  mv <- liftIO $ readIORef m
		    case mv of
			Open		-> return v
			Instantiated v	-> refresh (vApps v vs)
			    -- Note: it's important not to refresh the arguments to
			    -- instantiated meta variables (might lead to non-termination)

	    VAVDef x h vs	-> foldl vApp (vDef x h) <$> refresh vs
	    VAVCon x h ar vs	-> foldl vApp (vCon x h ar) <$> refresh vs
	    VAVProj v p vs	->
		do  v'  <- flip vProject p <$> refresh v
		    vs' <- refresh vs
		    return $ vApps v' vs'
	    VAVPi v cl		-> vPi <$> refresh v <*> refresh cl
	    VAVLam cl		-> vLam <$> refresh cl
	    VAVSort s		-> return v
	    VAVTop		-> return vTop
	    VAVPair v w		-> vPair <$> refresh v <*> refresh w
	    VAVSigma w f	-> vSigma <$> refresh w <*> refresh f
	    VAVStruct xs	-> vStruct <$> refresh xs

instance Refresh Type where
    refresh t = typeMapM refresh t

instance Refresh ContextEntry where
    refresh (Defined v t)   = Defined <$> refresh v <*> refresh t
    refresh (Abstract t)    = Abstract <$> refresh t

instance Refresh a => Refresh (Abs a) where
    refresh = absMapM refresh

instance Refresh ClauseValue where
    refresh (ClauseVal ps v) = ClauseVal ps <$> refresh v

instance Refresh DefValue where
    refresh (DefVal h ar fv i cs t) = DefVal h ar fv i <$> refresh cs <*> refresh t

instance Refresh Name where
    refresh x = return x

instance Refresh String where
    refresh x = return x

instance Refresh a => Refresh [a] where
    refresh xs = mapM refresh xs

instance (Refresh a, Refresh b) => Refresh (a,b) where
    refresh (x,y) = (,) <$> refresh x <*> refresh y


