
{--------------------------------------------------------------------------
    Logical constants
 --------------------------------------------------------------------------}

False : Type

(\/) : Type -> Type -> Type
X \/ Y = (C : Type) -> (X -> C) -> (Y -> C) -> C

orIntro1 : (X, Y : Type) -> X -> X \/ Y
orIntro1 X Y x C f g = f x

orIntro2 : (X, Y : Type) -> Y -> X \/ Y
orIntro2 X Y y C f g = g y

infixr 5 /\

(/\) : Type -> Type -> Type
X /\ Y = { X ; Y }

{--------------------------------------------------------------------------
    Predicates and relations
 --------------------------------------------------------------------------}

Pred : (A : Type) -> Type
Pred A = A -> Type

Rel : (A : Type) -> Type
Rel A = A -> A -> Type

Symmetric : (A : Type) -> Rel A -> Type
Symmetric A R = (x, y : A) -> R x y -> R y x

Transitive : (A : Type) -> Rel A -> Type
Transitive A R = (x, y, z : A) -> R x y -> R y z -> R x z

{--------------------------------------------------------------------------
    Equality
 --------------------------------------------------------------------------}

Equal : (A : Type) -> Rel A
Equal A x y = (P : A -> Type) -> P x -> P y

eqSym : (A : Type) -> Symmetric A (Equal A)
eqSym A x y xy P py = xy (\z -> P z -> P x) (\px -> px) py

eqTrans : (A : Type) -> Transitive A (Equal A)
eqTrans A x y z xy yz P px = yz P (xy P px)

{--------------------------------------------------------------------------
    Embeddings and well-founded relations
 --------------------------------------------------------------------------}

Field : (A : Type) -> Rel A -> Pred A
Field A R = \x -> { y : A; R x y } \/ { y : A; R y x }

EMB : (A : Type) -> Rel A -> (B : Type) -> Rel B -> Type
EMB A R B S =
  { !f	: A -> B
  ; !p1 : (x, y : A) -> Field A R x -> Field A R y -> R x y -> S (f x) (f y)
  ; !b	: B
  ; !p2	: (x : A) -> Field A R x -> S (f x) b
  }
	    -- The field requirements in p1 feel silly...
	    -- Aren't x and y automatically in the field
	    -- of R when we have R x y ?

field : (A : Type) -> (R : Rel A) -> (x, y : A) -> R x y ->
	 Field A R x /\ Field A R y
field A R x y xy = ( orIntro1 {z:A; R x z} {z:A; R z x} (y,xy)
		   , orIntro2 {z:A; R y z} {z:A; R z y} (x,xy)
		   )

WF : (A : Type) -> Rel A -> Type
WF A (<) =
    (P : Pred A) ->
    ( (x : A) -> 
      ( (y : A) -> y < x -> P y ) ->
      P x
    ) ->
    (x : A) -> P x

A : Type
R : Rel A

{--------------------------------------------------------------------------
    Universal system of notation
 --------------------------------------------------------------------------}

Morphism : (A : Type) -> Rel A -> (B : Type) -> Rel B -> Type
Morphism A R B S =
    { !f : A -> B
    ; !p : (x, y : A) -> R x y -> S (f x) (f y)
    }

IsUniversal : (A : Type) -> ((v : Type) -> Rel v -> A) -> Type
IsUniversal A i =
  (B : Type) -> (R : Rel B) -> (C : Type) -> (S : Rel C) ->
  Equal A (i B R) (i C S) -> Morphism B R C S

Universal : Type
Universal = { !A : Type
	    ; !i : (v : Type) -> Rel v -> A
	    ; !p : IsUniversal A i
	    }

{--------------------------------------------------------------------------
    Burali-Forti's paradox
 --------------------------------------------------------------------------}

wf0 : (u : Universal) -> Pred u.1
wf0 (A0,i0,_) x =
    { B : Type
    ; S : Rel B
    ; Equal A0 x (i0 B S) /\ WF B S
    }

emb : (u : Universal) -> Rel u.1
emb (A0,i0,_) x y =
    { A : Type
    ; R : Rel A
    ; B : Type
    ; S : Rel B
    ; Equal A0 x (i0 A R) /\ Equal A0 y (i0 B S) /\ EMB A R B S
    }

embTrans : (u : Universal) -> Transitive u.1 (emb u)
embTrans u x y z  (A1,R1,B1,S1,(xAR1,yBS1,embARBS1))
		  (A2,R2,B2,S2,(yAR2,zBS2,embARBS2)) =
    (A1,R1,B2,S2,(xAR1,zBS2,embARBS12))
  where
    m : Morphism B1 S1 A2 R2
    m = u.p B1 S1 A2 R2
	  (eqTrans u.A (u.i B1 S1) y (u.i A2 R2)
	      (eqSym u.A y (u.i B1 S1) yBS1)
	      yAR2
	  )

    h : A1 -> B1
    h = embARBS1.f

    gh : A1 -> A2
    gh a = m.f (h a)

    fgh : A1 -> B2
    fgh a = embARBS2.f (gh a)

    embARBS12 : EMB A1 R1 B2 S2
    embARBS12 =
      (fgh, \v w DR1v DR1w R1vw ->

	  let R2vw  = m.p (h v) (h w) (embARBS1.p1 v w DR1v DR1w R1vw)
	      DR2vw = field A2 R2 (gh v) (gh w) R2vw
	  in
	      embARBS2.p1 (gh v) (gh w)
		DR2vw.1 DR2vw.2
		R2vw
      , embARBS2.b
      , \v DR1v ->
	  embARBS2.p2 (gh v)
	    (field A2 R2 (gh v) (m.f embARBS1.b)
		(m.p (h v) embARBS1.b (embARBS1.p2 v DR1v))
	    ).1
      )

emb0 : (u : Universal) -> Rel u.1
emb0 u x y = emb u x y /\ wf0 u x /\ wf0 u y

lem1 : (u : Universal) -> WF u.1 (emb0 u)
lem1 u P ind x = ind x (\y  ( (A,R,B,S,yAR,xBS,(f,embRS,b,Sfb))
			    , (Ay,Ry,yARy,wfARy)
			    , (Bx,Sx,xBSx,wfBSx)
			    ) ->
    let
	iARARy = eqTrans u.A (u.i A R) y (u.i Ay Ry)
			     (eqSym u.A y (u.i A R) yAR) yARy
	iARyAR = eqSym u.A (u.i A R) (u.i Ay Ry) iARARy

	mARARy = u.p A R Ay Ry iARARy
	mARyAR = u.p Ay Ry A R iARyAR

	iBSBSx = eqTrans u.A (u.i B S) x (u.i Bx Sx)
			     (eqSym u.A x (u.i B S) xBS) xBSx
	iBSxBS = eqSym u.A (u.i B S) (u.i Bx Sx) iBSBSx

	mBSBSx = u.p B S Bx Sx iBSBSx
	mBSxBS = u.p Bx Sx B S iBSxBS

	fAAy : A -> Ay
	fAAy = mARARy.f

	fAyA : Ay -> A
	fAyA = mARyAR.f

	fBBx : B -> Bx
	fBBx = mBSBSx.f

	fBxB : Bx -> B
	fBxB = mBSxBS.f

    in	?   -- this is difficult...
  )

lem2 : (u : Universal) -> (A : Type) -> (R : Rel A) -> WF A R -> EMB A R u.1 (emb0 u)
--lem2 u A R wf = ?

lem3 : (A : Type) -> (R : Rel A) -> WF A R -> EMB A R A R -> False
--lem3 A R wf emb = ?

buraliForti : Universal -> False
buraliForti u = lem3 u.1 (emb0 u) (lem1 u) (lem2 u u.1 (emb0 u) (lem1 u))

{--------------------------------------------------------------------------
    Girard's paradox
 --------------------------------------------------------------------------}

A0 : Type
A0 = ((B : Type) -> Rel B -> Type) -> Type

i0 : (v : Type) -> Rel v -> A0
i0 B R x = x B R

univ0 : IsUniversal A0 i0
--univ0 = ?

false : False
false = buraliForti (A0, i0, univ0)

