
-- # import Property

-- backward compatibility
Eq : (A:Set) -> A -> A -> Prop
Eq A a b = a == b


-- The initial setup more or less follows Peter's suggestions.

-- The first order universe. (Is that correct terminology?)

Exp : Set

-- A binary function on terms.

app : Exp -> Exp -> Exp

-- Some nullary terms in this universe. 

Z  : Exp -- Zero
S0 : Exp -- Succ (the constructor)
P0 : Exp -- (+)  (the addition function)

-- Some handy abbreviations.

S : Exp -> Exp
S x = app S0 x
(+) : Exp -> Exp -> Exp
x + y = app (app P0 x) y

-- Predicate for total, finite natural numbers.

Nat : Exp -> Prop

-- axioms stating that the constructors are well typed

NZ : Nat Z
NS : (n : Exp) -> Nat n -> Nat (S n)

-- Induction on natural numbers

indNat : (P : Exp -> Prop)
          -> P Z
          -> ForAll {Exp} (\n -> Nat n ==> P n ==> P (S n))
          -> ForAll {Exp} (\n -> Nat n ==> P n)

-- Equality axioms representing the definition of addition.
-- I added the requirement that the expressions have to be natural
-- numbers. It's unclear if that is what we want.

plusZero : (m : Exp) -> Nat m ==> Eq Exp (m + Z) m
plusSucc : (m, n : Exp) -> Nat m ==> Nat n ==> Eq Exp (m + S n) (S (m + n))

-- The property we want to prove.

prop_add_commutative
  : ForAll {Exp} (\x -> Nat x ==>
       ForAll {Exp} (\y -> Nat y ==>
         Eq Exp (x + y) (y + x)
       )
     )
prop_add_commutative =
  indNat (\x -> ForAll {Exp} (\y -> Nat y ==> Eq Exp (x + y) (y + x)))
         (indNat (\y -> Eq Exp (Z + y) (y + Z))
                 (fol-plugin ())
                 (fol-plugin (plusZero, plusSucc, NZ, NS))
         )
         (\x xNat indHyp ->
           indNat (\y -> Eq Exp (S x + y) (y + S x))
                  (fol-plugin (plusZero, plusSucc, NZ, NS, xNat, indHyp))
                  (fol-plugin (plusZero, plusSucc, NZ, NS, xNat, indHyp))
         )
