We have to be careful with implicit arguments in the FOL translation.

--\begin{code}

-- # import Property
-- # import Bool
-- # import Nat

P (A:Set) : A -> Prop

axBool : (x:Bool) -> P x
axNot0 : Not (P 0)

unsound : true == false
unsound = fol-plugin (axBool, axNot0)
--\end{code}

But it seems that such a P is not definable in type theory 
(see attempts below).

Using parametricity, we can reason that 

  P _ = Q : Prop

for some fixed Q.  Then axBool and axNot0 are contradictory.

\begin{code}
T : Bool -> Set
T true  = Bool
T false = Nat

Q (x:Bool) : T x -> Prop
Q _ = if x True False  -- this feels bad, x should not be allowed here!
-- or:: Q = if x (\ b -> True) (\ n -> False)

-- Problem with meta variables:: T nat <= T ?  does not lead to inst ? = nat
qNat : (n:T false) -> (Q {false} n) -> False
qNat _ p = p

qNot0 : Not (Q {false} 0)
qNot0 = notI (\p -> p)

qBool : (b:T true) -> Q {true} b
qBool true  = tt
qBool false = tt

contra : False
contra = fol-plugin(qBool,qNot0)
\end{code}

Not definable::

Q (x:Bool) : T x -> Prop
Q true = True   --rejected!
Q false = True
Q 0 = False
Q (succ n) = False

Q (x:Bool) : T x -> Prop
Q a = a == true \/ a == false \/ Not (a == 0) 

An example with categories::

\begin{code}
Hom : Bool -> Bool -> Set
o (x,y,z:Bool) : Hom x y -> Hom y z -> Hom x z
id (x:Bool) : Hom x x

axId (x:Bool) : (f:Hom x (not x)) -> f `o` id == f /\ id `o` f == f

axInv (x:Bool) : (f:Hom x (not x)) -> (g:Hom (not x) x) -> f `o` g == id

-- f0 : Hom true false
-- f1 : Hom false true

assoc (w,x,y,z:Bool) : (f : Hom w x) -> (g : Hom x y) -> (h : Hom y z) -> 
  (f `o` g) `o` h == f `o` (g `o` h)
assoc = fol-plugin(axId,axInv)
-- at least Gandalf produces here a proof which does not type-check!!

collapse : (f,f':Hom true true) -> f == f'
collapse = fol-plugin (axId,axInv)
\end{code}

But we can have a non-collapsing model

  Hom x x       = Integer
  Hom x (not x) = {+1}
  id    = +1
  f o g = multiplication

But this model is not really definable in type theory, since we would
need sorts (subsets of types) here.
